@@ -1,21 +1,17 @@
 from __future__ import annotations
 
-import collections
 import dataclasses
-from typing import Dict, Optional, Tuple, Literal
-
-import numpy as np
-import pandas as pd
+import math
+from typing import Dict, Optional, Tuple, List
+from typing_extensions import Literal
 
+ON_DEMAND_OR_SPOT_VALUES: Tuple[OnDemandOrSpotType, OnDemandOrSpotType] = (
+    "on_demand",
+    "spot",
+)
 OnDemandOrSpotType = Literal["on_demand", "spot"]
 
 
-def assert_is_not_none(resources: Optional[Resources]) -> Resources:
-    """A helper for mypy"""
-    assert resources is not None
-    return resources
-
-
 @dataclasses.dataclass(frozen=True)
 class Resources:
     """
@@ -111,37 +107,63 @@ def remaining_resources_sort_key(
         return 1, None
 
 
+@dataclasses.dataclass(frozen=True)
+class CloudInstanceType:
+    name: str  # e.g. t2.micro
+    memory_gb: float  # e.g. 4 means 4 GiB
+    logical_cpu: int  # e.g. 2 means 2 logical (aka virtual) cpus
+    price: float  # 0.023 means 0.023 USD per hour to run the instance
+    # e.g. 0 for on-demand instances, >0 for spot instances, as a percentage, so values
+    # range from 0 to 100.
+    interruption_probability: float
+    on_demand_or_spot: OnDemandOrSpotType
+
+
+@dataclasses.dataclass
+class ChosenCloudInstanceType:
+    instance_type: CloudInstanceType
+    # e.g. 5 means we should allocate 5 of these instances. Will always be >1
+    num_instances: int
+
+    # The maximum number of workers that you could put on this instance type
+    workers_per_instance_full: int
+    # price per worker if we fit as many workers as possible onto the instance
+    price_per_worker_full: float
+
+    # for "internal" use:
+    workers_per_instance_current: int
+    price_per_worker_current: float
+
+
+@dataclasses.dataclass(frozen=True)
+class CloudInstance:
+    """Represents an instance launched by an InstanceRegistrar"""
+
+    public_dns_name: str
+    name: str
+
+    # TODO instance_type.interruption_probability should always use the latest data
+    # rather than always using the number from when the instance was launched
+    instance_type: ChosenCloudInstanceType
+
+
 def choose_instance_types_for_job(
     resources_required: Resources,
     num_workers_to_allocate: int,
     interruption_probability_threshold: float,
-    instance_types: pd.DataFrame,
-) -> pd.DataFrame:
+    original_instance_types: List[CloudInstanceType],
+) -> List[ChosenCloudInstanceType]:
     """
     This chooses how many of which instance types we should launch for a job with 1 or
-    more tasks where each task requires resources_required so that num_tasks_to_allocate
-    tasks can run in parallel. We choose the cheapest instances that have interruption
-    probability lower than or equal to the specified threshold. If you only want to use
-    on-demand instances that have 0 probability of interruption, you can set
-    interruption_probability_threshold to 0. If there are multiple instances that are
-    the cheapest, we choose the ones that have the lowest interruption probability. If
-    there are still multiple instances, then we diversify among those instance types (it
-    seems that interruptions are more likely to happen at the same time on the same
-    instance types).
-
-    instance_types should be a dataframe with these columns:
-    - instance_type: str, e.g. t2.micro
-    - memory_gb: float, e.g. 4 means 4 GiB
-    - logical_cpu: int, e.g. 2 means 2 logical (aka virtual) cpus
-    - price: float, e.g. 0.023 means 0.023 USD per hour to run the instance
-    - interruption_probability: float, e.g. 0 for on-demand instances, >0 for spot
-      instances, as a percentage, so values range from 0 to 100.
-    - on_demand_or_spot: str, "on_demand" or "spot" (not used in this function)
-
-    returns a dataframe with the same schema as instance_types, with additional columns:
-    - num_instances: e.g. 5 means we should allocate 5 of these instances
-    - workers_per_instance: how many workers should be able to run on that instance
-    Rows (i.e. instance types) will only be present if num_instances is at least 1.
+    more tasks where each task requires resources_required so that
+    num_workers_to_allocate tasks can run in parallel. We choose the cheapest instances
+    that have interruption probability lower than or equal to the specified threshold.
+    If you only want to use on-demand instances that have 0 probability of interruption,
+    you can set interruption_probability_threshold to 0. If there are multiple instances
+    that are the cheapest, we choose the ones that have the lowest interruption
+    probability. If there are still multiple instances, then we diversify among those
+    instance types (it seems that interruptions are more likely to happen at the same
+    time on the same instance types).
 
     TODO we should maybe have an option where e.g. if you want to allocate 53 workers
      worth of capacity for a 100-task job, it makes more sense to allocate e.g. 55 or 60
@@ -149,56 +171,83 @@ def choose_instance_types_for_job(
      workers of capacity
     """
 
-    # make a copy of instance_types, as we want to modify it in place
-    instance_types = instance_types.copy()
+    instance_types = []
 
-    # ignore anything with a higher interruption probability than what we want to
-    # tolerate
-    instance_types = instance_types[
-        instance_types["interruption_probability"] <= interruption_probability_threshold
-    ]
+    for orig_instance_type in original_instance_types:
+        # ignore anything with a higher interruption probability than what we want to
+        # tolerate
+        if (
+            orig_instance_type.interruption_probability
+            <= interruption_probability_threshold
+        ):
+            workers_per_instance_full = math.floor(
+                min(
+                    orig_instance_type.memory_gb / resources_required.memory_gb,
+                    orig_instance_type.logical_cpu / resources_required.logical_cpu,
+                )
+            )
+            # ignore instance types where we won't be able to fit even 1 worker
+            if workers_per_instance_full >= 1:
+                # if we get the maximum number of workers packed onto the instance type,
+                # what is our effective price per worker
+                price_per_worker_full = (
+                    orig_instance_type.price / workers_per_instance_full
+                )
 
-    # the maximum number of workers we could pack onto each instance type
-    instance_types["workers_per_instance"] = np.minimum(
-        np.floor(instance_types["memory_gb"] / resources_required.memory_gb),
-        np.floor(instance_types["logical_cpu"] / resources_required.logical_cpu),
-    ).astype(int)
-    # ignore instance types where we won't be able to fit even 1 worker
-    instance_types = instance_types[instance_types["workers_per_instance"] >= 1]
-    # if we get the maximum number of workers packed onto the instance type, what is our
-    # effective price per worker
-    instance_types["price_per_worker"] = (
-        instance_types["price"] / instance_types["workers_per_instance"]
-    )
+                instance_types.append(
+                    ChosenCloudInstanceType(
+                        orig_instance_type,
+                        0,
+                        workers_per_instance_full,
+                        price_per_worker_full,
+                        workers_per_instance_full,
+                        price_per_worker_full,
+                    )
+                )
 
-    instance_types = instance_types.reset_index(drop=True)
-    # This will keep track of how many of which instance type to use. E.g. {34: 5} tells
-    # us to use 5 instances of the instance_types.loc[34] instance type
-    instance_types_to_use: Dict[int, int] = collections.defaultdict(lambda: 0)
+    # no instance types can run even one worker
+    if len(instance_types) == 0:
+        return []
 
     while num_workers_to_allocate > 0:
-        best = instance_types.copy()
-
         # for larger instances, there might not be enough num_workers_to_allocate to
-        # make it "worth it" to use that larger instance because we won't have enough
+        # make it "worth it" to use that larger instance because we won't need enough
         # workers to fully pack the instance. So we recompute price_per_worker for those
         # instances assuming we only get to put num_workers_to_allocate on that
-        # instance.
-        too_large_index = (
-            instance_types["workers_per_instance"] > num_workers_to_allocate
-        )
-        best.loc[too_large_index, "price_per_worker"] = (
-            best.loc[too_large_index, "price"] / num_workers_to_allocate
-        )
+        # instance. We use ChosenCloudInstanceType.workers_per_instance_current and
+        # price_per_worker_current to keep track of this.
+        for instance_type in instance_types:
+            if instance_type.workers_per_instance_full > num_workers_to_allocate:
+                instance_type.price_per_worker_current = (
+                    instance_type.instance_type.price / num_workers_to_allocate
+                )
+                instance_type.workers_per_instance_current = num_workers_to_allocate
+            # as long as num_workers_to_allocate decreases on every iteration, we should
+            # never need to "reset" price_per_worker_current_iteration or
+            # workers_per_instance_current_iteration
 
         # Now find the instance types that have the lowest price per worker. If there
         # are multiple instance types that have the same price per worker (or are within
         # half a penny per hour), then take the ones that have the lowest probability of
         # interruption (within 1%)
         # TODO maybe the rounding should be configurable?
-        best = best[(best["price_per_worker"] - best["price_per_worker"].min()) < 0.005]
-        best = best[
-            (best["interruption_probability"] - best["interruption_probability"].min())
+        best_price_per_worker = min(
+            instance_type.price_per_worker_current for instance_type in instance_types
+        )
+        best = [
+            instance_type
+            for instance_type in instance_types
+            if instance_type.price_per_worker_current - best_price_per_worker < 0.005
+        ]
+        best_interruption_probability = min(
+            instance_type.instance_type.interruption_probability
+            for instance_type in best
+        )
+        best = [
+            instance_type
+            for instance_type in best
+            if instance_type.instance_type.interruption_probability
+            - best_interruption_probability
             < 1
         ]
 
@@ -212,12 +261,14 @@ def choose_instance_types_for_job(
         # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/spot-fleet-allocation-strategy.html
 
         # iterating from largest to smallest will make things easier
-        best = best.sort_values(["workers_per_instance"], ascending=False)
+        best.sort(
+            key=lambda instance_type: instance_type.workers_per_instance_full,
+            reverse=True,
+        )
 
         # take the first one no matter what
-        i = 0
-        instance_types_to_use[best.index[i]] += 1
-        num_workers_to_allocate -= best.loc[best.index[i], "workers_per_instance"]
+        best[0].num_instances += 1
+        num_workers_to_allocate -= best[0].workers_per_instance_current
 
         # Now that we've decreased num_workers_to_allocate, we need to make sure
         # price_per_worker is still accurate (i.e. num_workers_to_allocate could have
@@ -226,17 +277,23 @@ def choose_instance_types_for_job(
         # in a round-robin-ish fashion. Once there are no more instance types where
         # price_per_worker is still accurate, we'll go through the loop again and
         # recompute price_per_worker.
+        i = 0
         while True:
-            best = best[best["workers_per_instance"] <= num_workers_to_allocate]
+            best = [
+                instance_type
+                for instance_type in best
+                if instance_type.workers_per_instance_current <= num_workers_to_allocate
+            ]
             if len(best) == 0:
                 break
             # this is...very inexact because best is changing as we iterate, but the
             # idea is to walk through the options in best one by one
             i = (i + 1) % len(best)
-            instance_types_to_use[best.index[i]] += 1
-            num_workers_to_allocate -= best.loc[best.index[i], "workers_per_instance"]
+            best[i].num_instances += 1
+            num_workers_to_allocate -= best[i].workers_per_instance_current
 
-    # relies on defaultdict.keys() and .values() iterating in the same order
-    return instance_types.loc[instance_types_to_use.keys()].assign(
-        num_instances=instance_types_to_use.values()
-    )
+    return [
+        instance_type
+        for instance_type in instance_types
+        if instance_type.num_instances > 0
+    ]
